import { InvalidParam } from '@alfa-client/tech-shared';
import { existsAsHtmlElement, getElementComponentFromFixtureByCss, mock, Mock, triggerEvent } from '@alfa-client/test-utils';
import { ComponentFixture, TestBed } from '@angular/core/testing';
import { AbstractControl, FormControl, ReactiveFormsModule, UntypedFormGroup } from '@angular/forms';
import { expect } from '@jest/globals';
import { InfoIconComponent, TooltipDirective } from '@ods/system';
import { MockComponent, MockDirective } from 'ng-mocks';
import { of } from 'rxjs';
import { ValidationErrorComponent } from '../../../../../../design-component/src/lib/form/validation-error/validation-error.component';
import { getDataTestIdOf } from '../../../../../../tech-shared/test/data-test';
import { createInvalidParam } from '../../../../../../tech-shared/test/error';
import { createUserFormGroup } from '../../../../test/form';
import { UserFormService } from '../user.formservice';
import { UserFormRolesComponent } from './user-form-roles.component';

describe('UserFormRolesComponent', () => {
  let component: UserFormRolesComponent;
  let fixture: ComponentFixture<UserFormRolesComponent>;

  const validationErrorTestId: string = getDataTestIdOf('rollen-error');

  let formService: Mock<UserFormService>;

  beforeEach(async () => {
    formService = mock(UserFormService);

    await TestBed.configureTestingModule({
      imports: [UserFormRolesComponent, ReactiveFormsModule, MockComponent(InfoIconComponent), MockDirective(TooltipDirective)],
      providers: [{ provide: UserFormService, useValue: formService }],
    }).compileComponents();

    fixture = TestBed.createComponent(UserFormRolesComponent);
    component = fixture.componentInstance;
    component.formGroupParent = createUserFormGroup();
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  describe('component', () => {
    describe('ngOnInit', () => {
      it('should emit invalid params on invalid control', (done) => {
        const invalidParam: InvalidParam = createInvalidParam();
        const error: any = { dummy: invalidParam };
        const control: AbstractControl = new FormControl();
        component.formGroupParent = new UntypedFormGroup({ [UserFormService.CLIENT_ROLES]: control });

        component.ngOnInit();
        component.invalidParams$.subscribe((result) => {
          expect(result).toEqual([invalidParam]);
          done();
        });
        control.setErrors(error);
      });

      it('should emit empty array on valid control', (done) => {
        const control: AbstractControl = new FormControl();
        component.formGroupParent = new UntypedFormGroup({ [UserFormService.CLIENT_ROLES]: control });

        component.ngOnInit();
        component.invalidParams$.subscribe((result) => {
          expect(result).toEqual([]);
          done();
        });

        control.setErrors(null);
      });

      it('should mark as invalid', (done) => {
        const invalidParam: InvalidParam = createInvalidParam();
        const error: any = { dummy: invalidParam };
        const control: AbstractControl = new FormControl();
        component.formGroupParent = new UntypedFormGroup({ [UserFormService.CLIENT_ROLES]: control });

        component.ngOnInit();
        component.invalidParams$.subscribe(() => {
          expect(component.isValid).toBe(false);
          done();
        });

        control.setErrors(error);
      });

      it('should mark as valid', (done) => {
        const control: AbstractControl = new FormControl();
        component.formGroupParent = new UntypedFormGroup({ [UserFormService.CLIENT_ROLES]: control });

        component.ngOnInit();
        component.invalidParams$.subscribe(() => {
          expect(component.isValid).toBe(true);
          done();
        });

        control.setErrors(null);
      });
    });

    describe('handle alfa role change', () => {
      it('should call form service changeAlfaRole', () => {
        const formControlName: string = 'dummy';
        const value: boolean = true;

        component.handleAlfaRoleChange(formControlName, value);

        expect(formService.changeAlfaRole).toHaveBeenCalledWith(formControlName, value);
      });
    });

    describe('handle admin role change', () => {
      it('should call form service removeClientRolesValidationErrors', () => {
        component.handleAdminRoleChange();

        expect(formService.removeClientRolesValidationErrors).toHaveBeenCalled();
      });
    });
  });

  describe('template', () => {
    describe('validation error', () => {
      it('should exists', () => {
        existsAsHtmlElement(fixture, validationErrorTestId);
      });

      it('should have inputs', () => {
        const invalidParam: InvalidParam = createInvalidParam();
        component.invalidParams$ = of([invalidParam]);

        fixture.detectChanges();
        const validationErrorComponent: ValidationErrorComponent = getElementComponentFromFixtureByCss(
          fixture,
          validationErrorTestId,
        );

        expect(validationErrorComponent.invalidParams).toEqual([invalidParam]);
      });
    });

    describe('checkbox admin', () => {
      it('should call handleAdminRoleChange on inputChange emit', () => {
        component.handleAdminRoleChange = jest.fn();

        triggerEvent({
          fixture,
          elementSelector: getDataTestIdOf('checkbox-admin'),
          name: 'inputChange',
          data: true,
        });

        expect(component.handleAdminRoleChange).toHaveBeenCalled();
      });
    });

    describe('checkbox datenbeauftragung', () => {
      it('should call handleAdminRoleChange on inputChange emit', () => {
        component.handleAdminRoleChange = jest.fn();

        triggerEvent({
          fixture,
          elementSelector: getDataTestIdOf('checkbox-datenbeauftragung'),
          name: 'inputChange',
          data: true,
        });

        expect(component.handleAdminRoleChange).toHaveBeenCalled();
      });
    });

    describe('checkbox loeschen', () => {
      it('should call handleAlfaRoleChange on inputChange emit', () => {
        component.handleAlfaRoleChange = jest.fn();

        triggerEvent({
          fixture,
          elementSelector: getDataTestIdOf('checkbox-loeschen'),
          name: 'inputChange',
          data: true,
        });

        expect(component.handleAlfaRoleChange).toHaveBeenCalledWith(UserFormService.LOESCHEN, true);
      });
    });

    describe('checkbox user', () => {
      it('should call handleAlfaRoleChange on inputChange emit', () => {
        component.handleAlfaRoleChange = jest.fn();

        triggerEvent({
          fixture,
          elementSelector: getDataTestIdOf('checkbox-user'),
          name: 'inputChange',
          data: true,
        });

        expect(component.handleAlfaRoleChange).toHaveBeenCalledWith(UserFormService.USER, true);
      });
    });

    describe('checkbox poststelle', () => {
      it('should call handleAlfaRoleChange on inputChange emit', () => {
        component.handleAlfaRoleChange = jest.fn();

        triggerEvent({
          fixture,
          elementSelector: getDataTestIdOf('checkbox-poststelle'),
          name: 'inputChange',
          data: true,
        });

        expect(component.handleAlfaRoleChange).toHaveBeenCalledWith(UserFormService.POSTSTELLE, true);
      });
    });
  });
});
