/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { AdminOrganisationsEinheit, AdminOrganisationsEinheitService } from '@admin-client/organisations-einheit-shared';
import { ROUTES } from '@admin-client/shared';
import { AdminRoles, User, UserService } from '@admin-client/user-shared';
import {
  KeycloakErrorMessage,
  KeycloakFieldName,
  KeycloakFormService,
  KeycloakHttpErrorResponse,
  PatchConfig,
} from '@admin/keycloak-shared';
import { NavigationService } from '@alfa-client/navigation-shared';
import {
  checkBoxGroupsEmptyValidator,
  EMPTY_STRING,
  fieldEmptyValidator,
  fieldInvalidValidator,
  fieldLengthValidator,
  isLoaded,
  mapToResource,
  StateResource,
} from '@alfa-client/tech-shared';
import { SnackBarService } from '@alfa-client/ui';
import { Injectable } from '@angular/core';
import { AbstractControl, FormControl, UntypedFormBuilder, UntypedFormGroup } from '@angular/forms';
import { UrlSegment } from '@angular/router';
import { filter, Observable, take, tap } from 'rxjs';

@Injectable()
export class UserFormService extends KeycloakFormService<User> {
  public static readonly FIRST_NAME: string = 'firstName';
  public static readonly LAST_NAME: string = 'lastName';
  public static readonly USERNAME: string = 'username';
  public static readonly EMAIL: string = 'email';

  public static readonly CLIENT_ROLES: string = 'clientRoles';
  public static readonly GROUPS: string = 'groups';
  public static readonly ADMINISTRATION_GROUP: string = 'admin';
  public static readonly ADMIN: string = AdminRoles.ADMIN;
  public static readonly DATENBEAUFTRAGUNG: string = 'DATENBEAUFTRAGUNG';
  public static readonly ALFA_GROUP: string = 'alfa';
  public static readonly LOESCHEN: string = 'VERWALTUNG_LOESCHEN';
  public static readonly USER: string = 'VERWALTUNG_USER';
  public static readonly POSTSTELLE: string = 'VERWALTUNG_POSTSTELLE';

  _organisationsEinheitToGroupIdMap: Map<string, string> = new Map<string, string>();

  constructor(
    public formBuilder: UntypedFormBuilder,
    private userService: UserService,
    private adminOrganisationsEinheitService: AdminOrganisationsEinheitService,
    private navigationService: NavigationService,
    private snackBarService: SnackBarService,
  ) {
    super();
    this.init();
  }

  init() {
    this._initOrganisationsEinheiten().pipe(take(1)).subscribe();
  }

  _buildPatchConfig(url: UrlSegment[]): PatchConfig {
    const doPatch: boolean = this.shouldPatch(url);
    return {
      id: doPatch ? url[1].path : undefined,
      doPatch,
    };
  }

  private shouldPatch(url: UrlSegment[]): boolean {
    return url.length == 2 && url[0].path == 'benutzer' && url[1].path != 'neu';
  }

  _load(id: string): Observable<StateResource<User>> {
    return this.userService.getUserById(id);
  }

  _doAfterPatch(stateResource: StateResource<User>): void {
    this._updateAlfaRoleStates();
  }

  _initForm(): UntypedFormGroup {
    return this.formBuilder.group(
      {
        [UserFormService.FIRST_NAME]: new FormControl(EMPTY_STRING, fieldEmptyValidator(UserFormService.FIRST_NAME)),
        [UserFormService.LAST_NAME]: new FormControl(EMPTY_STRING, fieldEmptyValidator(UserFormService.LAST_NAME)),
        [UserFormService.USERNAME]: new FormControl(EMPTY_STRING, [fieldLengthValidator(UserFormService.USERNAME, 3, 255)]),
        [UserFormService.EMAIL]: new FormControl(EMPTY_STRING, [
          fieldInvalidValidator(UserFormService.EMAIL, /^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$/),
        ]),
        [UserFormService.CLIENT_ROLES]: this.formBuilder.group(
          {
            [UserFormService.ADMINISTRATION_GROUP]: this.formBuilder.group({
              [UserFormService.ADMIN]: new FormControl(false),
              [UserFormService.DATENBEAUFTRAGUNG]: new FormControl(false),
            }),
            [UserFormService.ALFA_GROUP]: this.formBuilder.group({
              [UserFormService.LOESCHEN]: new FormControl(false),
              [UserFormService.USER]: new FormControl(false),
              [UserFormService.POSTSTELLE]: new FormControl(false),
            }),
          },
          {
            validators: checkBoxGroupsEmptyValidator(UserFormService.CLIENT_ROLES, [
              UserFormService.ADMINISTRATION_GROUP,
              UserFormService.ALFA_GROUP,
            ]),
          },
        ),
        [UserFormService.GROUPS]: this.formBuilder.group({}),
      },
      { updateOn: 'submit' },
    );
  }

  _initOrganisationsEinheiten(): Observable<AdminOrganisationsEinheit[]> {
    const organisationsEinheitenGroup: UntypedFormGroup = this.getOrganisationsEinheitenGroup();
    return this.adminOrganisationsEinheitService.getAll().pipe(
      filter(isLoaded),
      mapToResource<AdminOrganisationsEinheit[]>(),
      tap((organisationsEinheiten: AdminOrganisationsEinheit[]): void => {
        this.setOrganisationsEinheitenIdsInMap(organisationsEinheiten);
        this._addOrganisationsEinheitenToForm(organisationsEinheiten, organisationsEinheitenGroup);
      }),
    );
  }

  private setOrganisationsEinheitenIdsInMap(organisationsEinheiten: AdminOrganisationsEinheit[]): void {
    organisationsEinheiten.forEach((organisationsEinheit: AdminOrganisationsEinheit) => {
      this._organisationsEinheitToGroupIdMap.set(organisationsEinheit.name, organisationsEinheit.id);
    });
  }

  _addOrganisationsEinheitenToForm(
    organisationsEinheiten: AdminOrganisationsEinheit[],
    organisationsEinheitenGroup: UntypedFormGroup,
  ): void {
    organisationsEinheiten.forEach((organisationsEinheit: AdminOrganisationsEinheit) => {
      organisationsEinheitenGroup.addControl(organisationsEinheit.name, new FormControl(false));
    });
  }

  public changeAlfaRole(formControlName: string, value: boolean) {
    this.setAlfaRole(formControlName, value);
    this.removeClientRolesValidationErrors();
    this._updateAlfaRoleStates();
  }

  private setAlfaRole(formControlName: string, value: boolean): void {
    this.getRoleGroup(UserFormService.ALFA_GROUP).get(formControlName).setValue(value, { emitEvent: false });
  }

  public removeClientRolesValidationErrors() {
    this.form.get(UserFormService.CLIENT_ROLES).setErrors(null);
  }

  _updateAlfaRoleStates(): void {
    if (this._isAnyAlfaRoleAssigned()) {
      this._disableAlfaRoles();
    } else {
      this._enableAlfaRoles();
    }
  }

  _isAnyAlfaRoleAssigned(): boolean {
    const alfaGroup: UntypedFormGroup = this.getRoleGroup(UserFormService.ALFA_GROUP);
    return Object.keys(alfaGroup.controls).some((key) => alfaGroup.controls[key].value);
  }

  _disableAlfaRoles(): void {
    for (const control of Object.values<AbstractControl>(this.getRoleGroup(UserFormService.ALFA_GROUP).controls)) {
      if (!control.value) this.disableControl(control);
    }
  }

  _enableAlfaRoles() {
    const alfaGroup: UntypedFormGroup = this.getRoleGroup(UserFormService.ALFA_GROUP);
    for (const control of Object.values<AbstractControl>(alfaGroup.controls)) {
      this.enableControl(control);
    }
  }

  private enableControl(control: AbstractControl): void {
    control.enable({ onlySelf: true });
  }

  private disableControl(control: AbstractControl): void {
    if (!control.value) control.disable({ onlySelf: true });
  }

  _doSubmit(): Observable<StateResource<User>> {
    const user: User = this._createUser();
    return this._createOrSave(user).pipe(
      tap((state: StateResource<User>): void => {
        if (!state.loading) this.handleOnCreateUserSuccess();
      }),
    );
  }

  _createOrSave(user: User): Observable<StateResource<User>> {
    if (this.isPatch()) {
      return this.userService.save({
        ...user,
        id: this._patchConfig.id,
        groupIds: user.groups.map((groupName: string): string => this._organisationsEinheitToGroupIdMap.get(groupName)),
      });
    } else {
      return this.userService.create(user);
    }
  }

  handleOnCreateUserSuccess(): void {
    this.snackBarService.showInfo(this.isPatch() ? 'Der Benutzer wurde gespeichert.' : 'Der Benutzer wurde angelegt.');
    this.navigationService.navigate(ROUTES.BENUTZER);
  }

  _createUser(): User {
    return {
      ...this._getFormValue(),
      enabled: true,
      clientRoles: {
        alfa: this._getRoles(UserFormService.ALFA_GROUP),
        admin: this._getRoles(UserFormService.ADMINISTRATION_GROUP),
      },
      groups: this._getActiveOrganisationsEinheiten(),
    };
  }

  _getRoles(roleGroup: string): string[] {
    return this.getActive(this.getRoleGroup(roleGroup));
  }

  _getActiveOrganisationsEinheiten(): string[] {
    return this.getActive(this.getOrganisationsEinheitenGroup());
  }

  private getActive(formGroup: UntypedFormGroup): string[] {
    return Object.keys(formGroup.controls).filter((key) => formGroup.get(key).value);
  }

  public getOrganisationsEinheitenGroup(): UntypedFormGroup {
    return <UntypedFormGroup>this.form.get(UserFormService.GROUPS);
  }

  private getRoleGroup(roleGroup: string): UntypedFormGroup {
    return <UntypedFormGroup>this.form.get(UserFormService.CLIENT_ROLES).get(roleGroup);
  }

  public getUserName(): string {
    return this.form.get(UserFormService.USERNAME).value;
  }

  _mapKeycloakErrorMessageToControlName(keycloakErrorMessage: KeycloakErrorMessage): string {
    switch (keycloakErrorMessage) {
      case KeycloakErrorMessage.USERNAME_IS_MISSING:
      case KeycloakErrorMessage.USERNAME_EXISTS:
        return UserFormService.USERNAME;
      case KeycloakErrorMessage.EMAIL_EXISTS:
        return UserFormService.EMAIL;
      default:
        throw new Error(`Cannot map keycloak error message ${keycloakErrorMessage} to control name`);
    }
  }

  _mapKeycloakFieldNameToControlName(keycloakFieldName: KeycloakFieldName): string {
    switch (keycloakFieldName) {
      case KeycloakFieldName.USERNAME:
        return UserFormService.USERNAME;
      case KeycloakFieldName.EMAIL:
        return UserFormService.EMAIL;
      default:
        throw new Error(`Cannot map keycloak field name ${keycloakFieldName} to control name`);
    }
  }

  _handleUnknownKeycloakError(error: KeycloakHttpErrorResponse): void {
    this.snackBarService.showError(
      this.isPatch() ? 'Der Benutzer konnte nicht gespeichert werden.' : 'Der Benutzer konnte nicht angelegt werden.',
    );
  }
}
